"""Test text plugin."""
from .. import util


class TestText(util.PluginTestCase):
    """Test text plugin."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: text
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.text:
                  convert_encoding: utf-8
            """
        ).format(self.tempdir)
        self.mktemp('.text.yml', config, 'utf-8')

    def test_text(self):
        """Test text."""

        bad_words = ['helo', 'begn']
        good_words = ['yes', 'word']
        self.mktemp('test.txt', '\n'.join(bad_words + good_words), 'utf-8')
        self.assert_spellcheck('.text.yml', bad_words)

    def test_text_utf16(self):
        """Test text `UTF-16`."""

        bad_words = ['helo', 'begn']
        good_words = ['yes', 'word']
        self.mktemp('test.txt', '\n'.join(bad_words + good_words), 'utf-16')
        self.assert_spellcheck('.text.yml', bad_words)

    def test_text_utf32(self):
        """Test text `UTF-32`."""

        bad_words = ['helo', 'begn']
        good_words = ['yes', 'word']
        self.mktemp('test.txt', '\n'.join(bad_words + good_words), 'utf-32')
        self.assert_spellcheck('.text.yml', bad_words)


class TestTextChained(util.PluginTestCase):
    """Test text extension."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: text
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.python:
              - pyspelling.filters.text:
                  convert_encoding: utf-8
            """
        ).format(self.tempdir)
        self.mktemp('.text.yml', config, 'utf-8')

    def test_text_after_python(self):
        """Test text after Python."""

        bad_words = ['helo', 'begn']
        good_words = ['yes', 'word']
        template = self.dedent(
            """
            \"""
            {}
            \"""
            """
        ).format('\n'.join(bad_words + good_words))
        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.text.yml', bad_words)



"""
Python filter.

Parse Python docstrings.
"""
from __future__ import unicode_literals
from .. import filters
import re
import textwrap
import tokenize
import codecs
import io
import unicodedata
import sys

F_SUPPORT = (3, 6) <= sys.version_info

tokenizer = tokenize.generate_tokens
PREV_DOC_TOKENS = (tokenize.INDENT, tokenize.DEDENT, tokenize.NEWLINE, tokenize.ENCODING)

RE_PY_ENCODE = re.compile(
    br'^[^\r\n]*?coding[:=]\s*([-\w.]+)|[^\r\n]*?\r?\n[^\r\n]*?coding[:=]\s*([-\w.]+)'
)

BACK_SLASH_TRANSLATION = {
    "\\a": '\a',
    "\\b": '\b',
    "\\f": '\f',
    "\\r": '\r',
    "\\t": '\t',
    "\\n": '\n',
    "\\v": '\v',
    "\\\\": '\\',
    '\\"': '"',
    "\\'": "'",
    "\\\n": ''
}

RE_ESC = re.compile(
    r'''(?x)
    (?P<special>\\['"abfrtnv\\\n])|
    (?P<char>\\U[\da-fA-F]{8}|\\u[\da-fA-F]{4}|\\x[\da-fA-F]{2})|
    (?P<oct>\\[0-7]{1,3})|
    (?P<name>\\N\{[^}{]*\})
    '''
)

RE_BESC = re.compile(
    r'''(?x)
    (?P<special>\\['"abfrtnv\\\n])|
    (?P<char>\\x[\da-fA-F]{2})|
    (?P<oct>\\[0-7]{1,3})
    '''
)

RE_FESC = re.compile(
    r'''(?x)
    (?P<special>\\['"abfrtnv\\\n])|
    (?P<char>\\U[\da-fA-F]{8}|\\u[\da-fA-F]{4}|\\x[\da-fA-F]{2})|
    (?P<oct>\\[0-7]{1,3})|
    (?P<name>\\N\{\{[^}{]*\}\})|
    (?P<fesc>\{\{)|
    (?P<format>\{[^}{]*\})
    '''
)

FE_RFESC = re.compile(
    r'''(?x)
    (?P<fesc>\{\{)|
    (?P<format>\{[^}{]*\})
    '''
)

RE_STRING_TYPE = re.compile(r'''((?:r|u|f|b)+)?(\'''|"""|'|")(.*?)\2''', re.I | re.S)

RE_NON_PRINTABLE = re.compile(r'[\x00-\x09\x0b-\x1f\x7f-\xff]+')

RE_VALID_STRING_TYPES = re.compile(r'^(?:\*|(?:[rubf]\*?)+)$', re.I)

RE_ITER_STRING_TYPES = re.compile(r'(\*|[rubf]\*?)', re.I)

FMT_STR = (
    'f', 'F',
    'fr', 'rf',
    'Fr', 'rF',
    'fR', 'Rf',
    'FR', 'RF'
)


class PythonFilter(filters.Filter):
    """Spelling Python."""

    MODULE = 0
    FUNCTION = 1
    CLASS = 2

    def __init__(self, options, default_encoding='utf-8'):
        """Initialization."""

        super().__init__(options, default_encoding)

    def get_default_config(self):
        """Get default configuration."""

        return {
            'comments': True,
            'docstrings': True,
            'strings': False,
            'group_comments': False,
            'string_types': 'fu',
            'decode_escapes': True
        }

    def setup(self):
        """Setup."""

        self.comments = self.config['comments']
        self.docstrings = self.config['docstrings']
        self.strings = self.config['strings']
        self.group_comments = self.config['group_comments']
        self.string_types, self.wild_string_types = self.eval_string_type(self.config['string_types'])
        self.decode_escapes = self.config['decode_escapes']

    def validate_options(self, k, v):
        """Validate options."""

        super().validate_options(k, v)
        if k == 'string_types':
            if RE_VALID_STRING_TYPES.match(v) is None:
                raise ValueError("{}: '{}' does not define valid string types".format(self.__class__.__name__, v))

    def header_check(self, content):
        """Special Python encoding check."""

        encode = None

        m = RE_PY_ENCODE.match(content)
        if m:
            if m.group(1):
                encode = m.group(1).decode('ascii')
            elif m.group(2):
                encode = m.group(2).decode('ascii')
        if encode is None:
            encode = 'utf-8'
        return encode

    def eval_string_type(self, text, is_string=False):
        """Evaluate string type."""

        stype = set()
        wstype = set()

        for m in RE_ITER_STRING_TYPES.finditer(text):
            value = m.group(0)
            if value == '*':
                wstype.add('u')
                wstype.add('f')
                wstype.add('r')
                wstype.add('b')
            elif value.endswith('*'):
                wstype.add(value[0].lower())
            else:
                stype.add(value.lower())

        if is_string and 'b' not in stype and 'f' not in stype:
            stype.add('u')

        return stype, wstype

    def get_string_type(self, text):
        """Get string type."""

        return self.eval_string_type(text, True)[0]

    def match_string(self, stype):
        """Match string type."""

        return not (stype - self.string_types) or bool(stype & self.wild_string_types)

    def replace_unicode(self, m):
        """Replace escapes."""

        groups = m.groupdict()
        esc = m.group(0)
        if groups.get('fesc'):
            value = m.group(0)
        elif groups.get('format'):
            value = ' '
        elif groups.get('special'):
            value = BACK_SLASH_TRANSLATION[esc]
        elif groups.get('char'):
            try:
                value = chr(int(esc[2:], 16))
            except Exception:
                value = esc
        elif groups.get('oct'):
            value = chr(int(esc[1:], 8))
        elif groups.get('name'):
            try:
                value = unicodedata.lookup(esc[3:-1])
            except Exception:
                value = esc
        return value.replace('\x00', '\n')

    def replace_bytes(self, m):
        """Replace escapes."""

        esc = m.group(0)
        value = esc
        if m.group('special'):
            value = BACK_SLASH_TRANSLATION[esc]
        elif m.group('char'):
            try:
                value = chr(int(esc[2:], 16))
            except Exception:
                value = esc
        elif m.group('oct'):
            value = int(esc[1:], 8)
            if value > 255:
                value -= 256
            value = chr(value)
        return value.replace('\x00', '\n')

    def process_strings(self, string, docstrings=False):
        """Process escapes."""

        m = RE_STRING_TYPE.match(string)
        stype = self.get_string_type(m.group(1) if m.group(1) else '')
        if not self.match_string(stype) and not docstrings:
            return '', False
        is_bytes = 'b' in stype
        is_raw = 'r' in stype
        is_format = 'f' in stype
        content = m.group(3)
        if is_raw and (not is_format or not self.decode_escapes):
            string = self.norm_nl(content)
        elif is_raw and is_format:
            string = self.norm_nl(FE_RFESC.sub(self.replace_unicode, content))
        elif is_bytes:
            string = self.norm_nl(RE_BESC.sub(self.replace_bytes, content))
        elif is_format:
            string = self.norm_nl(RE_FESC.sub(self.replace_unicode, content))
        else:
            string = self.norm_nl(RE_ESC.sub(self.replace_unicode, content))

        return textwrap.dedent(RE_NON_PRINTABLE.sub('\n', string) if is_bytes else string), is_bytes

    def _filter(self, text, context, encoding):
        """Retrieve the Python docstrings."""

        docstrings = []
        strings = []
        comments = []
        prev_token_type = tokenize.NEWLINE
        indent = ''
        name = None
        stack = [(context, 0, self.MODULE)]
        last_comment = False
        possible_fmt_str = None

        src = io.StringIO(text)

        for token in tokenizer(src.readline):
            token_type = token[0]
            value = token[1]
            line = str(token[2][0])
            line_num = token[2][0]

            value = token[1]
            line = str(token[2][0])

            # Track function and class ancestry
            if token_type == tokenize.NAME:
                possible_fmt_str = None
                if value in ('def', 'class'):
                    name = value
                elif name:
                    parent = stack[-1][2]
                    prefix = ''
                    if parent != self.MODULE:
                        prefix = '.' if parent == self.CLASS else ', '
                    if name == 'class':
                        stack.append(('%s%s' % (prefix, value), len(indent), self.CLASS))
                    elif name == 'def':
                        stack.append(('%s%s()' % (prefix, value), len(indent), self.FUNCTION))
                    name = None
                elif not F_SUPPORT and value in FMT_STR:
                    possible_fmt_str = (prev_token_type, value)
            elif token_type != tokenize.STRING:
                possible_fmt_str = None

            if token_type == tokenize.COMMENT and self.comments:
                # Capture comments
                if (
                    self.group_comments and
                    last_comment and
                    comments and (comments[-1][2] + 1) == line_num
                ):
                    # Group multiple consecutive comments
                    comments[-1][0] += '\n' + value[1:]
                    comments[-1][2] = line_num
                else:
                    if len(stack) > 1:
                        loc = "%s(%s): %s" % (stack[0][0], line, ''.join([crumb[0] for crumb in stack[1:]]))
                    else:
                        loc = "%s(%s)" % (stack[0][0], line)
                    comments.append([value[1:], loc, line_num])
            if token_type == tokenize.STRING:
                # Capture docstrings.
                # If we captured an `INDENT` or `NEWLINE` previously we probably have a docstring.
                # `NL` means end of line, but not the end of the Python code line (line continuation).
                if (
                    (prev_token_type in PREV_DOC_TOKENS) or
                    (possible_fmt_str and possible_fmt_str[0] in PREV_DOC_TOKENS)
                ):
                    if self.docstrings:
                        value = value.strip()
                        if possible_fmt_str and value.startswith(("'", "\"")):
                            value = possible_fmt_str[1] + value
                        string, is_bytes = self.process_strings(value, docstrings=True)
                        if string:
                            loc = "%s(%s): %s" % (stack[0][0], line, ''.join([crumb[0] for crumb in stack[1:]]))
                            docstrings.append(
                                filters.SourceText(string, loc, 'utf-8', 'py-docstring')
                            )
                elif self.strings:
                    value = value.strip()
                    if possible_fmt_str and value.startswith(("'", "\"")):
                        value = possible_fmt_str[1] + value
                    string, is_bytes = self.process_strings(value)
                    if string:
                        loc = "%s(%s): %s" % (stack[0][0], line, ''.join([crumb[0] for crumb in stack[1:]]))
                        strings.append(filters.SourceText(string, loc, 'utf-8', 'py-string'))

            if token_type == tokenize.INDENT:
                indent = value
            elif token_type == tokenize.DEDENT:
                indent = indent[:-4]
                if len(stack) > 1 and len(indent) <= stack[-1][1]:
                    stack.pop()

            # We purposefully avoid storing comments as comments can come before docstrings,
            # and that can mess up our logic. So if the token is a comment we won't track it,
            # and comments are always followed with `NL` so we ignore that as well.
            # We only care that docstrings are preceded by `NEWLINE`.
            if token_type != tokenize.COMMENT and not (last_comment and token_type == tokenize.NL):
                prev_token_type = token_type
                last_comment = False
            else:
                last_comment = True

        final_comments = []
        for comment in comments:
            final_comments.append(filters.SourceText(textwrap.dedent(comment[0]), comment[1], encoding, 'py-comment'))

        return docstrings + final_comments + strings

    def filter(self, source_file, encoding):  # noqa A001
        """Parse Python file returning content."""

        with codecs.open(source_file, 'r', encoding=encoding) as f:
            return self._filter(f.read(), source_file, encoding)

    def sfilter(self, source):
        """Filter."""

        return self._filter(source.text, source.context, source.encoding)


def get_plugin():
    """Return the filter."""

    return PythonFilter



"""Test stylesheets plugin."""
from .. import util


class TestStylesheetsCSS(util.PluginTestCase):
    """Test stylesheets CSS plugin."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.stylesheets:
                  stylesheets: css
                  group_comments: true
            """
        ).format(self.tempdir)
        self.mktemp('.stylesheets_css.yml', config, 'utf-8')

    def test_stylesheets_css(self):
        """Test stylesheets CSS."""

        bad_words = ['flga', 'graet']
        good_words = ['yes', 'word']
        template = self.dedent(
            """
            /*
            {}
            */
            p#id.class, p.other_id.class {{
                color: white;
            }}
            """
        ).format(
            '\n'.join(bad_words + good_words)
        )
        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.stylesheets_css.yml', bad_words)


class TestStylesheetsSCSS(util.PluginTestCase):
    """Test stylesheets SCSS plugin."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: scss
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.stylesheets:
                  stylesheets: scss
                  group_comments: true
            """
        ).format(self.tempdir)
        self.mktemp('.stylesheets_scss.yml', config, 'utf-8')

    def test_stylesheets_scss(self):
        """Test stylesheets SCSS."""

        bad_block = ['helo', 'begn']
        bad_comments = ['flga', 'graet']
        bad_comments2 = ['recieve', 'teh']
        bad_words = bad_block + bad_comments + bad_comments2
        good_words = ['yes', 'word']
        template = self.dedent(
            """
            /*
            {}
            */
            @mixin cover {{
              $color: red;
              // {}
              // {}
              @for $i from 1 through 5 {{
                &.bg-cover#{{$i}} {{ background-color: adjust-hue($color, 15deg * $i) }}
              }}
            }}
            .wrapper {{ @include cover }}
            """
        ).format(
            '\n'.join(bad_block + good_words),
            ' '.join(bad_comments + good_words),
            ' '.join(bad_comments2 + good_words)
        )
        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.stylesheets_scss.yml', bad_words)


class TestStylesheetsSASS(util.PluginTestCase):
    """Test stylesheets SASS plugin."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: scss
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.stylesheets:
                  stylesheets: sass
                  group_comments: true
            """
        ).format(self.tempdir)
        self.mktemp('.stylesheets_sass.yml', config, 'utf-8')

    def test_stylesheets_sass(self):
        """Test stylesheets SASS."""

        bad_block = ['helo', 'begn']
        bad_comments = ['flga', 'graet']
        bad_comments2 = ['recieve', 'teh']
        bad_words = bad_block + bad_comments + bad_comments2
        good_words = ['yes', 'word']
        template = self.dedent(
            """
            /*
            {}
            */
            =cover
              $color: red
              // {}
              // {}
              @for $i from 1 through 5
                &.bg-cover#{{$i}}
                  background-color: adjust-hue($color, 15deg * $i)
            .wrapper
              +cover
            """
        ).format(
            '\n'.join(bad_block + good_words),
            ' '.join(bad_comments + good_words),
            ' '.join(bad_comments2 + good_words)
        )
        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.stylesheets_sass.yml', bad_words)


class TestStylesheetsCSSChained(util.PluginTestCase):
    """Test chained stylesheets CSS plugin."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.text:
              - pyspelling.filters.stylesheets:
                  stylesheets: css
                  group_comments: true
            """
        ).format(self.tempdir)
        self.mktemp('.stylesheets_css.yml', config, 'utf-8')

    def test_stylesheets_css_after_text(self):
        """Test stylesheets CSS."""

        bad_words = ['flga', 'graet']
        good_words = ['yes', 'word']
        template = self.dedent(
            """
            /*
            {}
            */
            p#id.class, p.other_id.class {{
                color: white;
            }}
            """
        ).format(
            '\n'.join(bad_words + good_words)
        )
        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.stylesheets_css.yml', bad_words)



"""Test CPP plugin."""
from .. import util


class TestCPP(util.PluginTestCase):
    """Test CPP plugin."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: cpp
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.cpp:
                  group_comments: true
                  trigraphs: true
            """
        ).format(self.tempdir)
        self.mktemp('.cpp.yml', config, 'utf-8')

    def test_cpp(self):
        """Test CPP."""

        bad_block = ['helo', 'begn']
        bad_comments = ['flga', 'graet']
        bad_comments2 = ['recieve', 'teh']
        bad_comments3 = ['hpoe', 'lvoe']
        bad_words = bad_block + bad_comments + bad_comments2 + bad_comments3
        good_words = ['yes', 'word']
        template = self.dedent(
            """
            /*
            {}
            */
            uint8_t func() {{
                uint8_t tsdd = 3;
                // {}
                // {} \
                reurn tsdd;

                uint32_t trigraph_test = (CONSTANT_1 ??' ADKASLD) ??' CONSTANT_2; // {}

                rtuern ddst;
            }}
            """
        ).format(
            '\n'.join(bad_block + good_words),
            ' '.join(bad_comments + good_words),
            ' '.join(bad_comments2 + good_words),
            ' '.join(bad_comments3 + good_words)
        )

        # Capture comment continuation
        bad_words.extend(['reurn', 'tsdd'])
        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.cpp.yml', bad_words)


class TestCPPGroupedComments(util.PluginTestCase):
    """Test CPP plugin."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: cpp
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.cpp:
                  group_comments: true
                  trigraphs: true
              - pyspelling.filters.context:
                  context_visible_first: true
                  escapes: '\\[\\`]'
                  delimiters:
                  # Ignore multiline content between fences (fences can have 3 or more back ticks)
                  # ```
                  # content
                  # ```
                  - open: '(?s)^(?P<open> *`{{3,}})$'
                    close: '^(?P=open)$'
                  # Ignore text between inline back ticks
                  - open: '(?P<open>`+)'
                    close: '(?P=open)'
            """
        ).format(self.tempdir)
        self.mktemp('.cpp.yml', config, 'utf-8')

    def test_cpp_grouped_comments(self):
        """Test CPP."""

        template = self.dedent(
            """
            uint8_t func() {

              uint32_t test = 3; // what
              // bbbb

              // ```
              // akdjal dlkajdlf dkkjkj
              // skjdl djkjf lslsl
              // ```

              // aaaa
            }
            """
        )
        bad_words = ['aaaa', 'bbbb']
        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.cpp.yml', bad_words)


class TestCPPStringAllow(util.PluginTestCase):
    """Test CPP allow filter."""

    def setup_fs(self):
        """Setup."""

        template = self.dedent(
            r"""
            uint8_t func() {
                auto s0 =   "aaaa"; // char
                auto s1 =  L"bbbb"; // wchar_t
                auto s2 = u8"cccc"; // char
                auto s3 =  u"dddd"; // char16_t
                auto s4 =  U"eeee"; // char32_t
                auto R0 =   R"("ffff")"; // const char*
                auto R1 =   R"delim("gggg")delim";    // const char*
                auto R3 =  LR"("hhhh")"; // const wchar_t*
                auto R4 = u8R"("iiii")"; // const char*, encoded as UTF-8
                auto R5 =  uR"("jjjj")"; // const char16_t*, encoded as UTF-16
                auto R6 =  UR"("kkkk")"; // const char32_t*, encoded as UTF-32
            }
            """
        )

        self.mktemp('test.txt', template, 'utf-8')

    def test_exclude_raw(self):
        """Test raw exclusion."""

        config = self.dedent(
            """
            matrix:
            - name: cpp
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.cpp:
                  strings: true
                  line_comments: false
                  group_comments: true
                  string_types: uls
            """
        ).format(self.tempdir)
        self.mktemp('.cpp.yml', config, 'utf-8')
        self.assert_spellcheck('.cpp.yml', ['aaaa', 'bbbb', 'cccc', 'dddd', 'eeee'])

    def test_exclude_unicode(self):
        """Test Unicode exclusion."""

        config = self.dedent(
            """
            matrix:
            - name: cpp
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.cpp:
                  strings: true
                  line_comments: false
                  group_comments: true
                  string_types: rls
            """
        ).format(self.tempdir)
        self.mktemp('.cpp.yml', config, 'utf-8')
        self.assert_spellcheck('.cpp.yml', ["aaaa", "bbbb", "ffff", "gggg", "hhhh"])

    def test_exclude_wide(self):
        """Test wide exclusion."""

        config = self.dedent(
            """
            matrix:
            - name: cpp
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.cpp:
                  strings: true
                  line_comments: false
                  group_comments: true
                  string_types: rus
            """
        ).format(self.tempdir)
        self.mktemp('.cpp.yml', config, 'utf-8')
        self.assert_spellcheck('.cpp.yml', ["aaaa", "cccc", "dddd", "eeee", "ffff", "gggg", "iiii", "jjjj", "kkkk"])

    def test_exclude_standard(self):
        """Test standard exclusion."""

        config = self.dedent(
            """
            matrix:
            - name: cpp
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.cpp:
                  strings: true
                  line_comments: false
                  group_comments: true
                  string_types: rul
            """
        ).format(self.tempdir)
        self.mktemp('.cpp.yml', config, 'utf-8')
        self.assert_spellcheck('.cpp.yml', ["bbbb", "cccc", "dddd", "eeee", "hhhh", "iiii", "jjjj", "kkkk"])

    def test_include_all(self):
        """Test standard exclusion."""

        config = self.dedent(
            """
            matrix:
            - name: cpp
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.cpp:
                  strings: true
                  line_comments: false
                  group_comments: true
                  string_types: '*'
            """
        ).format(self.tempdir)
        self.mktemp('.cpp.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.cpp.yml',
            ["aaaa", "bbbb", "cccc", "dddd", "eeee", "ffff", "gggg", "hhhh", "iiii", "jjjj", "kkkk"]
        )

    def test_include_all_raw(self):
        """Test standard exclusion."""

        config = self.dedent(
            """
            matrix:
            - name: cpp
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.cpp:
                  strings: true
                  line_comments: false
                  group_comments: true
                  string_types: 'r*'
            """
        ).format(self.tempdir)
        self.mktemp('.cpp.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.cpp.yml',
            ["ffff", "gggg", "hhhh", "iiii", "jjjj", "kkkk"]
        )

    def test_include_all_unicode(self):
        """Test standard exclusion."""

        config = self.dedent(
            """
            matrix:
            - name: cpp
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.cpp:
                  strings: true
                  line_comments: false
                  group_comments: true
                  string_types: 'u*'
            """
        ).format(self.tempdir)
        self.mktemp('.cpp.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.cpp.yml',
            ["cccc", "dddd", "eeee", "iiii", "jjjj", "kkkk"]
        )


class TestCPPStrings(util.PluginTestCase):
    """Test CPP plugin."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: cpp
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.cpp:
                  strings: true
                  line_comments: false
                  group_comments: true
                  string_types: ruls
            """
        ).format(self.tempdir)
        self.mktemp('.cpp.yml', config, 'utf-8')

    def test_cpp_strings(self):
        """Test CPP strings."""

        bad_char = ['helo', 'begn']
        bad_wchar = ['flga', 'graet']
        bad_u8 = ['recieve', 'teh']
        bad_u16 = ['acept', 'thruogh']
        bad_u32 = ['hpoe', 'lvoe']
        bad_raw = ['aaaa', 'bbbb']
        bad_raw_delim = ['cccc', 'dddd']
        bad_raw_wide = ['eeee', 'ffff']
        bad_raw_u8 = ['gggg', 'hhhh']
        bad_raw_u16 = ['iiii', 'jjjj']
        bad_raw_u32 = ['kkkk', 'llll']

        bad_words = (
            bad_char + bad_wchar + bad_u8 + bad_u16 + bad_u32 + bad_raw + bad_raw_wide +
            bad_raw_delim + bad_raw_u8 + bad_raw_u16 + bad_raw_u32
        )
        good_words = ['yes', 'word']
        template = self.dedent(
            r"""
            uint8_t func() {{
                auto c0 =   'xxxx'; // char
                auto s0 =   "{}"; // char
                auto s1 =  L"{}"; // wchar_t
                auto s2 = u8"{}"; // char
                auto s3 =  u"{}"; // char16_t
                auto s4 =  U"{}"; // char32_t
                auto R0 =   R"("{}")"; // const char*
                auto R1 =   R"delim("{}")delim";    // const char*
                auto R3 =  LR"("{}")"; // const wchar_t*
                auto R4 = u8R"("{}")"; // const char*, encoded as UTF-8
                auto R5 =  uR"("{}")"; // const char16_t*, encoded as UTF-16
                auto R6 =  UR"("{}")"; // const char32_t*, encoded as UTF-32
            }}
            """
        ).format(
            ' '.join(bad_char + good_words),
            ' '.join(bad_wchar + good_words),
            ' '.join(bad_u8 + good_words),
            ' '.join(bad_u16 + good_words),
            ' '.join(bad_u32 + good_words),
            ' '.join(bad_raw + good_words),
            ' '.join(bad_raw_delim + good_words),
            ' '.join(bad_raw_wide + good_words),
            ' '.join(bad_raw_u8 + good_words),
            ' '.join(bad_raw_u16 + good_words),
            ' '.join(bad_raw_u32 + good_words)
        )

        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.cpp.yml', bad_words)

    def test_cpp_strings_escapes(self):
        """Test CPP strings escapes."""

        template = self.dedent(
            r"""
            uint8_t func() {{
                auto s0 =   "\x61gggg\141 \ntext\
                             \u0061yyyy \\ubbbb"; // char
                auto s1 =  L"\x0061hhhh \ntexts"; // wchar_t
                auto s2 = u8"\141\x000000061\u0061\U00000061"; // char
                auto s3 =  u"\142\x000000062\u0062\U00000062"; // char16_t
                auto s4 =  U"\143\x000000063\u0063\U00000063"; // char32_t
                auto R0 =   R"("\xaa \nbad")"; // const char*
                auto R1 =   R"delim("\xbb
                                     \vbad")delim";    // const char*
                auto R3 =  LR"("\xcc")"; // const wchar_t*
                auto R4 = u8R"("\xdd")"; // const char*, encoded as UTF-8
                auto R5 =  uR"("\xee")"; // const char16_t*, encoded as UTF-16
                auto R6 =  UR"("\xff")"; // const char32_t*, encoded as UTF-32
            }}
            """
        )

        # `char*` and `wchar*` will have escapes `\x`, `\u`, and `\U` stripped out as we don't know the encoding.
        # Unicode strings will have all escapes decoded.
        # Raw will decode none.
        bad_words = [
            'agggga', 'ahhhh', 'ayyyy', 'aaaa', 'bbbb', 'cccc',
            'xaa', 'xbb', 'xcc', 'xdd', 'xee', 'xff',
            'ubbbb', 'nbad', 'vbad'
        ]

        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.cpp.yml', bad_words)


class TestCPPGeneric(util.PluginTestCase):
    """Test CPP plugin."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: cpp
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.cpp:
                  group_comments: true
                  generic_mode: true
            """
        ).format(self.tempdir)
        self.mktemp('.cpp.yml', config, 'utf-8')

    def test_cpp(self):
        """Test CPP."""

        bad_block = ['helo', 'begn']
        bad_comments = ['flga', 'graet']
        bad_comments2 = ['recieve', 'teh']
        bad_words = bad_block + bad_comments + bad_comments2
        good_words = ['yes', 'word']
        template = self.dedent(
            r"""
            /*
            {}
            */
            uint8_t func() {{
                uint8_t tsdd = 3;
                // {}
                // {} \
                reurn tsdd;

                rtuern ddst;
            }}
            """
        ).format(
            '\n'.join(bad_block + good_words),
            ' '.join(bad_comments + good_words),
            ' '.join(bad_comments2 + good_words)
        )

        # Capture comment continuation
        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.cpp.yml', bad_words)


class TestCPPChained(util.PluginTestCase):
    """Test chained CPP plugin."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: cpp
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.text:
              - pyspelling.filters.cpp:
                  group_comments: true
            """
        ).format(self.tempdir)
        self.mktemp('.cpp.yml', config, 'utf-8')

    def test_cpp_after_text(self):
        """Test CPP after text."""

        bad_block = ['helo', 'begn']
        bad_comments = ['flga', 'graet']
        bad_comments2 = ['recieve', 'teh']
        bad_words = bad_block + bad_comments + bad_comments2
        good_words = ['yes', 'word']
        template = self.dedent(
            """
            /*
            {}
            */
            uint8_t func() {{
                uint8_t tsdd = 3;
                // {}
                // {}
                reurn tsdd;
            }}
            """
        ).format(
            '\n'.join(bad_block + good_words),
            ' '.join(bad_comments + good_words),
            ' '.join(bad_comments2 + good_words)
        )
        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.cpp.yml', bad_words)



"""Base plugin class."""
from collections import OrderedDict
import warnings


class Plugin:
    """Base plugin."""

    def __init__(self, options):
        """Initialization."""

        self.config = self.get_default_config()
        if self.config is None:
            warnings.warn(
                "'{}' did not provide a default config. ".format(self.__class__.__name__) +
                "All plugins in the future should provide a default config.",
                category=FutureWarning,
                stacklevel=1
            )
            self.config = options
        else:
            self.override_config(options)
        self.setup()

    def get_default_config(self):
        """Get default configuration."""

        return None

    def setup(self):
        """Setup."""

    def override_config(self, options):
        """Override the default configuration."""

        for k, v in options.items():
            # Reject names not in the default configuration
            if k not in self.config:
                raise KeyError("'{}' is not a valid option for '{}'".format(k, self.__class__.__name__))
            self.validate_options(k, v)
            self.config[k] = v

    def validate_options(self, k, v):
        """Validate options."""

        args = [self.__class__.__name__, k]
        # Booleans
        if isinstance(self.config[k], bool) and not isinstance(v, bool):
            raise ValueError("{}: option '{}' must be a bool type.".format(*args))
        # Strings
        elif isinstance(self.config[k], str) and not isinstance(v, str):
            raise ValueError("{}: option '{}' must be a str type.".format(*args))
        # Integers (whole floats allowed)
        elif (
            isinstance(self.config[k], int) and
            (not isinstance(v, int) and not (isinstance(v, float) and v.is_integer()))
        ):
            raise ValueError("{}: option '{}' must be a int type.".format(*args))
        # Floats (integers allowed)
        elif isinstance(self.config[k], float) and not isinstance(v, (int, float)):
            raise ValueError("{}: option '{}' must be a float type.".format(*args))
        # Basic iterables (list, tuple, sets)
        elif isinstance(self.config[k], (list, tuple, set)) and not isinstance(v, list):
            raise ValueError("{}: option '{}' must be a float type.".format(*args))
        # Dictionaries
        elif isinstance(self.config[k], (dict, OrderedDict)) and not isinstance(v, (dict, OrderedDict)):
            raise ValueError("{}: option '{}' must be a dict type.".format(*args))

    def reset(self):
        """Reset anything needed on each iteration."""



"""Test Python plugin."""
from .. import util


class TestPython(util.PluginTestCase):
    """Test Python plugin."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: python
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.python:
                  group_comments: true
            """
        ).format(self.tempdir)
        self.mktemp('.python.yml', config, 'utf-8')

    def test_python(self):
        """Test Python."""

        bad_docstring = ['helo', 'begn']
        bad_comments = ['flga', 'graet']
        bad_comments2 = ['recieve', 'teh']
        bad_words = bad_docstring + bad_comments + bad_comments2
        good_words = ['yes', 'word']
        template = self.dedent(
            """
            \"""
            {}
            \"""
            def function():
                # {}
                # {}
            """
        ).format(
            '\n'.join(bad_docstring + good_words),
            ' '.join(bad_comments + good_words),
            ' '.join(bad_comments2 + good_words)
        )
        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.python.yml', bad_words)


class TestPythonStrings(util.PluginTestCase):
    """Test Python plugin."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: python
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.python:
                  strings: true
                  string_types: bfru
                  group_comments: true
            """
        ).format(self.tempdir)
        self.mktemp('.pystrings.yml', config, 'utf-8')

    def test_python_continue(self):
        """Test strings with line continuation."""

        bad_words = ['helo', 'begn']
        good_words = ['yes', 'word']
        template = self.dedent(
            r"""
            def function():
                test = "{} \tthe\
                aaaa"
            """
        ).format(
            ' '.join(bad_words + good_words)
        )
        bad_words.append('aaaa')
        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.pystrings.yml', bad_words)

    def test_python_raw(self):
        """Test Python raw strings."""

        bad_words = ['helo', 'begn']
        good_words = ['yes', 'word']
        template = self.dedent(
            r"""
            def function():
                test = r"{} \tthe"
            """
        ).format(
            ' '.join(bad_words + good_words)
        )
        bad_words.append('tthe')
        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.pystrings.yml', bad_words)

    def test_python_unicode(self):
        """Test Python Unicode strings."""

        bad_words = ['helo', 'begn']
        good_words = ['yes', 'word']
        template = self.dedent(
            r"""
            def function():
                test = "{} \141\x61\u0061\U00000061\N{{LATIN SMALL LETTER A}}"
            """
        ).format(
            ' '.join(bad_words + good_words)
        )
        bad_words.append('aaaaa')
        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.pystrings.yml', bad_words)

    def test_python_bytes(self):
        """Test Python byte strings."""

        bad_words = ['helo', 'begn']
        good_words = ['yes', 'word']
        template = self.dedent(
            r"""
            def function():
                test = b"\x03\x03\x02{} \tthe \141\541\141\541\x03\xff\x03\x02"
            """
        ).format(
            '\\x03'.join(bad_words + good_words)
        )
        bad_words.append('aaaa')
        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.pystrings.yml', bad_words)

    def test_python_raw_bytes(self):
        """Test Python raw byte strings."""

        bad_words = ['helo', 'begn']
        good_words = ['yes', 'word']
        template = self.dedent(
            r"""
            def function():
                test = rb"{} \tthe \xff"
            """
        ).format(
            ' '.join(bad_words + good_words)
        )
        bad_words.extend(['tthe', 'xff'])
        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.pystrings.yml', bad_words)

    def test_python_format(self):
        """Test Python format strings."""

        bad_words = ['helo', 'begn']
        good_words = ['yes', 'word']
        template = self.dedent(
            r"""
            def function():
                aaaa = "text"
                test = f"{} \tthe {{aaaa}}"
            """
        ).format(
            ' '.join(bad_words + good_words)
        )

        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.pystrings.yml', bad_words)

    def test_python_raw_format(self):
        """Test Python raw format strings."""

        bad_words = ['helo', 'begn']
        good_words = ['yes', 'word']
        template = self.dedent(
            r"""
            def function():
                aaaa = "text"
                test = fr"{} \tthe {{aaaa}}"
            """
        ).format(
            ' '.join(bad_words + good_words)
        )

        bad_words.append('tthe')

        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.pystrings.yml', bad_words)


class TestPythonStringAllow(util.PluginTestCase):
    """Test Python allow filter."""

    def setup_fs(self):
        """Setup."""

        template = self.dedent(
            r"""
            def func():
                s0 =   "aaaa"
                s1 =  u"bbbb"
                s2 =  b"cccc"
                s3 =  f"dddd"
                s4 =  r"eeee"
                s5 =  br"ffff"
                s7 =  fr"gggg"
            """
        )

        self.mktemp('test.txt', template, 'utf-8')

    def test_exclude_unicode(self):
        """Test exclude Unicode."""

        config = self.dedent(
            """
            matrix:
            - name: python
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.python:
                  strings: true
                  string_types: brf
            """
        ).format(self.tempdir)
        self.mktemp('.python.yml', config, 'utf-8')
        self.assert_spellcheck('.python.yml', ['cccc', 'dddd', 'ffff', 'gggg'])

    def test_exclude_bytes(self):
        """Test exclude bytes."""

        config = self.dedent(
            """
            matrix:
            - name: python
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.python:
                  strings: true
                  string_types: urf
            """
        ).format(self.tempdir)
        self.mktemp('.python.yml', config, 'utf-8')
        self.assert_spellcheck('.python.yml', ['aaaa', 'bbbb', 'dddd', 'eeee', 'gggg'])

    def test_exclude_format(self):
        """Test exclude format."""

        config = self.dedent(
            """
            matrix:
            - name: python
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.python:
                  strings: true
                  string_types: bur
            """
        ).format(self.tempdir)
        self.mktemp('.python.yml', config, 'utf-8')
        self.assert_spellcheck('.python.yml', ['aaaa', 'bbbb', 'cccc', 'eeee', 'ffff'])

    def test_exclude_raw(self):
        """Test exclude raw."""

        config = self.dedent(
            """
            matrix:
            - name: python
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.python:
                  strings: true
                  string_types: buf
            """
        ).format(self.tempdir)
        self.mktemp('.python.yml', config, 'utf-8')
        self.assert_spellcheck('.python.yml', ['aaaa', 'bbbb', 'cccc', 'dddd'])

    def test_include_all(self):
        """Test include all."""

        config = self.dedent(
            """
            matrix:
            - name: python
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.python:
                  strings: true
                  string_types: '*'
            """
        ).format(self.tempdir)
        self.mktemp('.python.yml', config, 'utf-8')
        self.assert_spellcheck('.python.yml', ['aaaa', 'bbbb', 'cccc', 'dddd', 'eeee', 'ffff', 'gggg'])

    def test_include_all_raw(self):
        """Test include all raw."""

        config = self.dedent(
            """
            matrix:
            - name: python
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.python:
                  strings: true
                  string_types: 'r*'
            """
        ).format(self.tempdir)
        self.mktemp('.python.yml', config, 'utf-8')
        self.assert_spellcheck('.python.yml', ['eeee', 'ffff', 'gggg'])

    def test_include_all_unicode(self):
        """Test include all Unicode."""

        config = self.dedent(
            """
            matrix:
            - name: python
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.python:
                  strings: true
                  string_types: 'u*'
            """
        ).format(self.tempdir)
        self.mktemp('.python.yml', config, 'utf-8')
        self.assert_spellcheck('.python.yml', ['aaaa', 'bbbb', 'eeee'])


class TestPythonChained(util.PluginTestCase):
    """Test Python plugin."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: python
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.text:
              - pyspelling.filters.python:
                  group_comments: true
            """
        ).format(self.tempdir)
        self.mktemp('.python.yml', config, 'utf-8')

    def test_python_after_text(self):
        """Test Python after text."""

        bad_docstring = ['helo', 'begn']
        bad_comments = ['flga', 'graet']
        bad_comments2 = ['recieve', 'teh']
        bad_words = bad_docstring + bad_comments + bad_comments2
        good_words = ['yes', 'word']
        template = self.dedent(
            """
            \"""
            {}
            \"""
            def function():
                # {}
                # {}
            """
        ).format(
            '\n'.join(bad_docstring + good_words),
            ' '.join(bad_comments + good_words),
            ' '.join(bad_comments2 + good_words)
        )
        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.python.yml', bad_words)



"""Test utilities."""
import contextlib
import os
import shutil
import unittest
import warnings
from textwrap import dedent
from pyspelling import spellcheck
import sys

# Below is general helper stuff that Python uses in `unittests`.  As these
# not meant for users, and could change without notice, include them
# ourselves so we aren't surprised later.
TESTFN = '@test'
WIN = sys.platform.startswith('win')
HUNSPELL = 'hunspell.exe' if WIN else 'hunspell'
ASPELL = 'aspell.exe' if WIN else 'aspell'

# Disambiguate `TESTFN` for parallel testing, while letting it remain a valid
# module name.
TESTFN = "{}_{}_tmp".format(TESTFN, os.getpid())


def which(executable):
    """See if executable exists."""

    location = None
    if os.path.basename(executable) != executable:
        if os.path.isfile(executable):
            location = executable
    else:
        paths = [x for x in os.environ["PATH"].split(os.pathsep) if not x.isspace()]
        paths.extend([x for x in os.environ.get("TOX_SPELL_PATH", "").split(os.pathsep) if not x.isspace()])
        for path in paths:
            exe = os.path.join(path, executable)
            if os.path.isfile(exe):
                location = exe
                break
    return location


@contextlib.contextmanager
def change_cwd(path, quiet=False):
    """
    Return a context manager that changes the current working directory.

    Arguments:
      path: the directory to use as the temporary current working directory.
      quiet: if False (the default), the context manager raises an exception
        on error.  Otherwise, it issues only a warning and keeps the current
        working directory the same.
    """

    saved_dir = os.getcwd()
    try:
        os.chdir(path)
    except OSError:
        if not quiet:
            raise
        warnings.warn('tests may fail, unable to change CWD to: ' + path,
                      RuntimeWarning, stacklevel=3)
    try:
        yield os.getcwd()
    finally:
        os.chdir(saved_dir)


class PluginTestCase(unittest.TestCase):
    """Test plugin."""

    def dedent(self, content):
        """
        Normalize indentation of text.

        When normalizing indentation of text, you often leave a new line at the start, so strip it as well.
        """

        return dedent(content).replace('\n', '', 1)

    def mktemp(self, filename, content, encoding):
        """Make temp directory."""

        filename = os.path.join(self.tempdir, os.path.normpath(filename))
        base, file = os.path.split(filename)
        if not os.path.exists(base):
            retry = 3
            while retry:
                try:
                    os.makedirs(base)
                    retry = 0
                except Exception:
                    retry -= 1
        with open(filename, "wb") as f:
            f.write(content.encode(encoding))

    def setUp(self):
        """Setup."""

        self.tempdir = TESTFN + "_dir"
        self.setup_fs()

    def setup_fs(self):
        """Setup file system (common files used across multiple tests)."""

    def tearDown(self):
        """Cleanup."""

        retry = 3
        while retry:
            try:
                shutil.rmtree(self.tempdir)
                retry = 0
            except Exception:
                retry -= 1

    def assert_spell_required(self, running):
        """Check if what we are running matches what we request."""

        required = os.environ.get("TOX_SPELL_REQUIRE", "any")

        if required not in ("both", "any", "aspell", "hunspell"):
            raise RuntimeError("Invalid value of '{}' in 'TOX_SPELL_REQUIRE'".format(required))

        if running not in ("both", "aspell", "hunspell"):
            raise RuntimeError(
                "Required tests are not being run (currently running '{}'). ".format(running) +
                "Make sure spell checker can be found and " +
                "'TOX_SPELL_REQUIRE' is set appropriately (currently '{}')".format(required)
            )

        if required != running and (required != "any" and running != "both"):
            raise RuntimeError(
                "'TOX_SPELL_REQUIRE' env variable, which is "
                "'{}', is not compatible with '{}'".format(running, required)
            )

    def assert_spellcheck(self, config_file, expected, names=None, groups=None, sources=None, verbose=4):
        """Spell check."""

        hunspell_location = which(HUNSPELL)
        aspell_location = which(ASPELL)
        if hunspell_location and aspell_location:
            running = "both"
        elif hunspell_location:
            running = "hunspell"
        elif aspell_location:
            running = "aspell"
        else:
            running = "none"
        self.assert_spell_required(running)

        if hunspell_location:
            words = set()
            for results in spellcheck(
                os.path.join(self.tempdir, config_file),
                names=names,
                groups=groups,
                sources=sources,
                checker='hunspell',
                binary=hunspell_location,
                debug=True,
                verbose=verbose
            ):
                if results.error:
                    print(results.error)
                words |= set(results.words)
            self.assertEqual(sorted(expected), sorted(list(words)))
        if aspell_location:
            words = set()
            for results in spellcheck(
                os.path.join(self.tempdir, config_file),
                names=names,
                groups=groups,
                sources=sources,
                checker='aspell',
                binary=aspell_location,
                debug=True,
                verbose=verbose
            ):
                if results.error:
                    print(results.error)
                words |= set(results.words)
            self.assertEqual(sorted(expected), sorted(list(words)))



"""
XML filter.

Detect encoding from XML header.
"""
from __future__ import unicode_literals
from .. import filters
import re
import codecs
import bs4
import soupsieve as sv
from collections import deque

NON_CONTENT = (bs4.Doctype, bs4.Declaration, bs4.CData, bs4.ProcessingInstruction)

RE_XML_START = re.compile(
    b'^(?:(' +
    b'<\\?xml[^>]+?>' +  # ASCII like
    b')|(' +
    re.escape('<?xml'.encode('utf-32-be')) + b'.+?' + re.escape('>'.encode('utf-32-be')) +
    b')|(' +
    re.escape('<?xml'.encode('utf-32-le')) + b'.+?' + re.escape('>'.encode('utf-32-le')) +
    b')|(' +
    re.escape('<?xml'.encode('utf-16-be')) + b'.+?' + re.escape('>'.encode('utf-16-be')) +
    b')|(' +
    re.escape('<?xml'.encode('utf-16-le')) + b'.+?' + re.escape('>'.encode('utf-16-le')) +
    b'))'
)

RE_XML_ENCODE = re.compile(br'''(?i)^<\?xml[^>]*encoding=(['"])(.*?)\1[^>]*\?>''')
RE_XML_ENCODE_U = re.compile(r'''(?i)^<\?xml[^>]*encoding=(['"])(.*?)\1[^>]*\?>''')


class XmlFilter(filters.Filter):
    """Spelling Python."""

    default_capture = ['*|*']

    break_tags = set()

    def __init__(self, options, default_encoding='utf-8'):
        """Initialization."""

        super().__init__(options, default_encoding)

    def get_default_config(self):
        """Get default configuration."""

        return {
            "comments": True,
            "attributes": [],
            "break_tags": [],
            "ignores": [],
            "captures": self.default_capture,
            "namespaces": {}
        }

    def setup(self):
        """Setup."""

        self.user_break_tags = set(self.config['break_tags'])
        self.comments = self.config['comments']
        self.attributes = set(self.config['attributes'])
        self.parser = 'xml'
        self.type = 'xml'
        ignores = ','.join(self.config['ignores'])
        self.ignores = sv.compile(ignores, self.config['namespaces']) if ignores.strip() else None
        captures = ','.join(self.config['captures'])
        self.captures = sv.compile(captures, self.config['namespaces']) if captures.strip() else None

    def _has_xml_encode(self, content):
        """Check XML encoding."""

        encode = None

        m = RE_XML_START.match(content)
        if m:
            if m.group(1):
                m2 = RE_XML_ENCODE.match(m.group(1))

                if m2:
                    enc = m2.group(2).decode('ascii')

                    try:
                        codecs.getencoder(enc)
                        encode = enc
                    except LookupError:
                        pass
            else:
                if m.group(2):
                    enc = 'utf-32-be'
                    text = m.group(2)
                elif m.group(3):
                    enc = 'utf-32-le'
                    text = m.group(3)
                elif m.group(4):
                    enc = 'utf-16-be'
                    text = m.group(4)
                elif m.group(5):
                    enc = 'utf-16-le'
                    text = m.group(5)
                try:
                    m2 = RE_XML_ENCODE_U.match(text.decode(enc))
                except Exception:  # pragma: no cover
                    m2 = None

                if m2:
                    enc = m2.group(2)

                    try:
                        codecs.getencoder(enc)
                        encode = enc
                    except Exception:
                        pass

        return encode

    def header_check(self, content):
        """Special HTML encoding check."""

        return self._has_xml_encode(content)

    def is_break_tag(self, el):
        """Check if tag is an element we should break on."""

        name = el.name
        return name in self.break_tags or name in self.user_break_tags

    def store_blocks(self, el, blocks, text, force_root):
        """Store the text as desired."""

        if force_root or el.parent is None or self.is_break_tag(el):
            content = ''.join(text)
            if content:
                blocks.append((content, self.construct_selector(el)))
            text = []
        return text

    def construct_selector(self, el, attr=''):
        """Construct an selector for context."""

        selector = deque()

        ancestor = el
        while ancestor and ancestor.parent:
            if ancestor is not el:
                selector.appendleft(ancestor.name)
            else:
                tag = ancestor.name
                prefix = ancestor.prefix
                sel = ''
                if prefix:
                    sel += prefix + '|'
                sel = tag
                if attr:
                    sel += '[%s]' % attr
                selector.appendleft(sel)
            ancestor = ancestor.parent
        return '>'.join(selector)

    def extract_tag_metadata(self, el):
        """Extract meta data."""

    def reset(self):
        """Reset."""

    def to_text(self, tree, force_root=False):
        """
        Extract text from tags.

        Skip any selectors specified and include attributes if specified.
        Ignored tags will not have their attributes scanned either.
        """

        self.extract_tag_metadata(tree)

        text = []
        attributes = []
        comments = []
        blocks = []

        if not (self.ignores.match(tree) if self.ignores else None):
            # The root of the document is the BeautifulSoup object
            capture = self.captures.match(tree) if self.captures is not None else None
            # Check attributes for normal tags
            if capture:
                for attr in self.attributes:
                    value = tree.attrs.get(attr, '').strip()
                    if value:
                        sel = self.construct_selector(tree, attr=attr)
                        attributes.append((value, sel))

            # Walk children
            for child in tree.children:
                string = str(child).strip()
                is_comment = isinstance(child, bs4.Comment)
                if isinstance(child, bs4.element.Tag):
                    t, b, a, c = self.to_text(child)
                    text.extend(t)
                    attributes.extend(a)
                    comments.extend(c)
                    blocks.extend(b)
                # Get content if not the root and not a comment (unless we want comments).
                elif not isinstance(child, NON_CONTENT) and (not is_comment or self.comments):
                    string = str(child).strip()
                    if string:
                        if is_comment:
                            sel = self.construct_selector(tree) + '<!--comment-->'
                            comments.append((string, sel))
                        elif capture:
                            text.append(string)
                            text.append(' ')
        elif self.comments:
            for child in tree.descendants:
                if isinstance(child, bs4.Comment):
                    string = str(child).strip()
                    if string:
                        sel = self.construct_selector(tree) + '<!--comment-->'
                        comments.append((string, sel))

        text = self.store_blocks(tree, blocks, text, force_root)

        if tree.parent is None or force_root:
            return blocks, attributes, comments
        else:
            return text, blocks, attributes, comments

    def _filter(self, text, context, encoding):
        """Filter the source text."""

        content = []
        blocks, attributes, comments = self.to_text(bs4.BeautifulSoup(text, self.parser))
        if self.comments:
            for c, desc in comments:
                content.append(filters.SourceText(c, context + ': ' + desc, encoding, self.type + 'comment'))
        if self.attributes:
            for a, desc in attributes:
                content.append(filters.SourceText(a, context + ': ' + desc, encoding, self.type + 'attribute'))
        for b, desc in blocks:
            content.append(filters.SourceText(b, context + ': ' + desc, encoding, self.type + 'content'))
        return content

    def filter(self, source_file, encoding):  # noqa A001
        """Parse XML file."""

        with codecs.open(source_file, 'r', encoding=encoding) as f:
            text = f.read()
        return self._filter(text, source_file, encoding)

    def sfilter(self, source):
        """Filter."""

        return self._filter(source.text, source.context, source.encoding)


def get_plugin():
    """Return the filter."""

    return XmlFilter



"""Test JavaScript plugin."""
from .. import util


class TestJavaScript(util.PluginTestCase):
    """Test JavaScript plugin."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: javascript
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.javascript:
                  jsdocs: true
                  group_comments: true
            """
        ).format(self.tempdir)
        self.mktemp('.javascript.yml', config, 'utf-8')

    def test_javascript(self):
        """Test CPP."""

        bad_docstring = ['helo', 'begn']
        bad_block = ['adn', 'highight']
        bad_comments = ['flga', 'graet']
        bad_comments2 = ['recieve', 'teh']
        bad_words = bad_docstring + bad_block + bad_comments + bad_comments2
        good_words = ['yes', 'word']
        template = self.dedent(
            """
            /**
             * {}
             */
            function Func(arg, arg2) {{
                /* {} */

                var x = 3;
                // {}
                // {}
            }}
            """
        ).format(
            ' '.join(bad_docstring + good_words),
            ' '.join(bad_block + good_words),
            ' '.join(bad_comments + good_words),
            ' '.join(bad_comments2 + good_words)
        )
        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.javascript.yml', bad_words)


class TestJavaScriptStrings(util.PluginTestCase):
    """Test JavaScript plugin."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: javascript
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.javascript:
                  jsdocs: true
                  strings: true
                  group_comments: true
            """
        ).format(self.tempdir)
        self.mktemp('.javascript.yml', config, 'utf-8')

    def test_javascript_strings(self):
        """Test JavaScript strings."""

        bad_string1 = ['helo', 'begn']
        bad_string2 = ['adn', 'highight']
        bad_words = bad_string1 + bad_string2
        good_words = ['yes', 'word']
        template = self.dedent(
            r"""
            function Func(arg, arg2) {{

              test = "{} \
              {}";

              test2 = "\141\x61\u0061\u{{00000061}}\a";
            }}
            """
        ).format(
            ' '.join(bad_string1 + good_words),
            ' '.join(bad_string2 + good_words)
        )
        bad_words.append('aaaaa')
        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.javascript.yml', bad_words)

    def test_javascript_temp_strings(self):
        """Test JavaScript."""

        bad_string1 = ['helo', 'begn']
        bad_string2 = ['adn', 'highight']
        bad_words = bad_string1 + bad_string2
        good_words = ['yes', 'word']
        template = self.dedent(
            r"""
            function Func(arg, arg2) {{

              test = `{} \
              {}`;

              xxxx = 3;

              test2 = `\x61\u0061\u{{00000061}}\a ${{ {{"test": `value ${{xxxx}} yyyy`}}["test"] }} no`;
            }}
            """
        ).format(
            ' '.join(bad_string1 + good_words),
            ' '.join(bad_string2 + good_words)
        )
        bad_words.extend(['aaaa', 'yyyy'])
        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.javascript.yml', bad_words)


class TestJavaScriptGroupedComments(util.PluginTestCase):
    """Test JavaScript plugin."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: javascript
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.javascript:
                  jsdocs: true
                  strings: true
                  group_comments: true
              - pyspelling.filters.context:
                  context_visible_first: true
                  escapes: '\\[\\`]'
                  delimiters:
                  # Ignore multiline content between fences (fences can have 3 or more back ticks)
                  # ```
                  # content
                  # ```
                  - open: '(?s)^(?P<open> *`{{3,}})$'
                    close: '^(?P=open)$'
                  # Ignore text between inline back ticks
                  - open: '(?P<open>`+)'
                    close: '(?P=open)'
            """
        ).format(self.tempdir)
        self.mktemp('.javascript.yml', config, 'utf-8')

    def test_javascript_grouped_comments(self):
        """Test JavaScript grouped comments."""

        template = self.dedent(
            r"""
            function Func(arg, arg2) {{

              var test = 3; // what
              // bbbb

              // ```
              // akdjal dlkajdlf dkkjkj
              // skjdl djkjf lslsl
              // ```

              // aaaa
            }}
            """
        )
        bad_words = ['aaaa', 'bbbb']
        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.javascript.yml', bad_words)


class TestJavaScriptChained(util.PluginTestCase):
    """Test chained JavaScript plugin."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: javascript
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.text:
              - pyspelling.filters.javascript:
                  jsdocs: true
                  group_comments: true
            """
        ).format(self.tempdir)
        self.mktemp('.javascript.yml', config, 'utf-8')

    def test_javascript_after_text(self):
        """Test JavaScript after text."""

        bad_docstring = ['helo', 'begn']
        bad_block = ['adn', 'highight']
        bad_comments = ['flga', 'graet']
        bad_comments2 = ['recieve', 'teh']
        bad_words = bad_docstring + bad_block + bad_comments + bad_comments2
        good_words = ['yes', 'word']
        template = self.dedent(
            """
            /**
             * {}
             */
            function Func(arg, arg2) {{
                /* {} */

                var x = 3;
                // {}
                // {}
            }}
            """
        ).format(
            ' '.join(bad_docstring + good_words),
            ' '.join(bad_block + good_words),
            ' '.join(bad_comments + good_words),
            ' '.join(bad_comments2 + good_words)
        )
        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.javascript.yml', bad_words)
